import json
import logging
import threading
import time
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Union

from .apis.maxim import MaximAPI
from .cache.cache import MaximCache
from .cache.inMemory import MaximInMemoryCache
from .filterObjects import (IncomingQuery, QueryObject, findAllMatches,
                            findBestMatch, parseIncomingQuery)
from .logger.logger import Logger, LoggerConfig
from .models.dataset import DatasetEntry
from .models.folder import Folder, FolderEncoder
from .models.prompt import (Prompt, PromptData, PromptVersion, RuleGroupType,
                            RuleType, VersionAndRulesWithPromptIdEncoder,
                            VersionsAndRules, VersionSpecificDeploymentConfig)
from .models.queryBuilder import QueryRule

maximLogger = logging.getLogger("MaximSDK")

@dataclass
class Config():
    apiKey: str
    baseUrl: Optional[str] = None
    cache: Optional[MaximCache] = None
    debug: Optional[bool] = False


EntityType = {
    "PROMPT": "PROMPT",
    "FOLDER": "FOLDER"
}


class Maxim:
    def __init__(self, config: Config):        
        maximLogger.setLevel(logging.DEBUG if config.debug else logging.INFO)        
        self.baseUrl = config.baseUrl or "https://app.getmaxim.ai"
        self.apiKey = config.apiKey
        self.is_running = True
        self.mutex = threading.Lock()
        self.__isDebug = config.debug or False
        self.__cache = config.cache or MaximInMemoryCache()
        self.__loggers: Dict[str, Logger] = {}
        self.__syncThread = threading.Timer(60, self.__syncEntities)
        self.__syncThread.start()
        self.__syncPrompts()
        self.__syncFolders()        

    def __syncEntities(self):
        maximLogger.debug(" Syncing prompts and folders")
        if not self.is_running:
            return
        with self.mutex:
            self.__syncThread = threading.Timer(60, self.__syncEntities)
            self.__syncThread.start()
            self.__syncPrompts()
            self.__syncFolders()
            maximLogger.debug("Syncing completed")

    def __syncPrompts(self):
        maximLogger.debug("Syncing prompts")
        try:
            prompts = MaximAPI.getPrompts(self.baseUrl, self.apiKey)
            maximLogger.debug(f"Found {len(prompts)} prompts")
            for prompt in prompts:
                try:
                    self.__cache.set(self.__getCacheKey(
                        EntityType['PROMPT'], prompt.promptId), json.dumps(prompt, cls=VersionAndRulesWithPromptIdEncoder))
                except Exception as err:
                    maximLogger.error(err)
        except Exception as err:
            maximLogger.error(f"Error while syncing prompts: {err}")

    def __syncFolders(self):
        maximLogger.debug("Syncing folders")
        try:
            folders = MaximAPI.getFolders(self.baseUrl, self.apiKey)
            maximLogger.debug(f"Found {len(folders)} folders")
            for folder in folders:
                try:
                    self.__cache.set(self.__getCacheKey(
                        EntityType['FOLDER'], folder.id), json.dumps(folder, cls=FolderEncoder))
                except Exception as err:
                    maximLogger.error(err)
        except Exception as err:
            maximLogger.error(f"Error while syncing folders: {err}")

    def __getCacheKey(self, entity: str, id: str) -> str:
        if entity == 'PROMPT':
            return f"prompt:{id}"
        else:
            return f"folder:{id}"

    def __getPromptFromCache(self, key: str) -> Optional[VersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        if 'promptId' in json_data:
            json_data.pop('promptId')
        return VersionsAndRules.from_dict(json_data)

    def __getAllPromptsFromCache(self) -> Optional[List[VersionsAndRules]]:
        keys = self.__cache.getAllKeys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("prompt:")]
        promptList = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if 'promptId' in json_data:
                    json_data.pop('promptId')
                promptList.append(VersionsAndRules.from_dict(json_data))
        return promptList

    def __getFolderFromCache(self, key: str) -> Optional[Folder]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return Folder(**json_data)

    def __getAllFoldersFromCache(self) -> Optional[List[Folder]]:
        keys = self.__cache.getAllKeys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("folder:")]
        return [Folder(**json.loads(d)) for d in data if d is not None]

    def __formatPrompt(self, prompt: PromptVersion) -> Prompt:
        json_data = {
            'promptId': prompt.promptId,
            'version': prompt.version,
            'versionId': prompt.id,
            'messages': prompt.config.messages if prompt.config else None,
            'modelParameters': prompt.config.modelParameters if prompt.config else None
        }
        return Prompt(**json_data)

    def __getPromptVersionForRule(self, promptVersionAndRules: VersionsAndRules, rule: Optional[QueryRule] = None) -> Optional[Prompt]:
        if rule:
            incomingQuery = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
            objects = []
            for versionId, versionRules in promptVersionAndRules.rules.items():
                for versionRule in versionRules:
                    if not versionRule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                    if not version:
                        continue
                    query = versionRule.rules.query
                    if version.config and version.config.tags:
                        parsedIncomingQuery = parseIncomingQuery(
                            incomingQuery.query)
                        tags = version.config.tags
                        for key, value in tags.items():
                            if value == None:
                                continue
                            if parsedIncomingQuery is None:
                                continue
                            incomingQueryFields = [
                                q.field for q in parsedIncomingQuery]
                            if key in incomingQueryFields:
                                query.rules.append(
                                    RuleType(field=key, operator='=', value=value))
                    objects.append(QueryObject(query=query, id=versionId))

            deployedVersionObject = findBestMatch(objects, incomingQuery)
            if deployedVersionObject:
                deployedVersion = next(
                    (v for v in promptVersionAndRules.versions if v.id == deployedVersionObject.id), None)
                if deployedVersion:
                    return self.__formatPrompt(deployedVersion)

        else:
            if promptVersionAndRules.rules:
                for versionId, versionRules in promptVersionAndRules.rules.items():
                    logging.debug(f"type: {type(versionRules[0].rules)}")
                    doesQueryExist = any(
                        ruleElm.rules.query != None for ruleElm in versionRules)
                    if doesQueryExist:
                        deployedVersion = next(
                            (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                        if deployedVersion:
                            return self.__formatPrompt(deployedVersion)
            else:
                return self.__formatPrompt(promptVersionAndRules.versions[0])

        if promptVersionAndRules.fallbackVersion:
            maximLogger.info(
                f"************* FALLBACK TRIGGERED : VERSION : {promptVersionAndRules.fallbackVersion} *************")
            return self.__formatPrompt(promptVersionAndRules.fallbackVersion)
        return None

    def __getFoldersForRule(self, folders: List[Folder], rule: QueryRule) -> List[Folder]:
        incomingQuery = IncomingQuery(
            query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
        objects = []
        for folder in folders:
            query = RuleGroupType(rules=[], combinator='AND')
            if not folder.tags:
                continue
            parsedIncomingQuery = parseIncomingQuery(incomingQuery.query)
            tags = folder.tags
            for key, value in tags.items():
                if key in [q.field for q in parsedIncomingQuery]:
                    # if not isinstance(value,None):
                    query.rules.append(
                        RuleType(field=key, operator='=', value=value))
            if not query.rules:
                continue
            objects.append(QueryObject(query=query, id=folder.id))

        folderObjects = findAllMatches(objects, incomingQuery)
        ids = [fo.id for fo in folderObjects]
        return [f for f in folders if f.id in ids]

    def getPrompt(self, id: str, rule: Optional[QueryRule] = None) -> Optional[Prompt]:
        key = self.__getCacheKey('PROMPT', id)
        versionAndRules = self.__getPromptFromCache(key)
        if versionAndRules is None:
            versionAndRules = MaximAPI.getPrompt(self.baseUrl, self.apiKey, id)
            if len(versionAndRules.versions) == 0:
                return None
            self.__cache.set(id, json.dumps(versionAndRules))
        if not versionAndRules:
            return None
        prompt = self.__getPromptVersionForRule(versionAndRules, rule)
        if not prompt:
            return None
        return prompt

    def getPrompts(self, rule: QueryRule) -> List[Prompt]:
        versionAndRules = self.__getAllPromptsFromCache()
        prompts = []
        if versionAndRules is None or len(versionAndRules) == 0:
            self.__syncEntities()
            versionAndRules = self.__getAllPromptsFromCache()
        if not versionAndRules:
            return []
        for v in versionAndRules:
            if rule.scopes:
                if 'folder' not in rule.scopes.keys():
                    return []
                else:
                    if rule.scopes['folder'] != v.folderId:
                        continue
            prompt = self.__getPromptVersionForRule(v, rule)
            if prompt != None:
                prompts.append(prompt)
        if len(prompts) == 0:
            return []
        return prompts

    def getFolderById(self, id: str) -> Optional[Folder]:
        key = self.__getCacheKey('FOLDER', id)
        folder = self.__getFolderFromCache(key)
        if folder is None:
            try:
                folder = MaximAPI.getFolder(self.baseUrl, self.apiKey, id)
                if not folder:
                    return None
                self.__cache.set(key, json.dumps(folder))
            except Exception as e:
                return None
        return folder

    def getFolders(self, rule: QueryRule) -> List[Folder]:
        folders = self.__getAllFoldersFromCache()
        if folders is None or len(folders) == 0:
            self.__syncEntities()
            folders = self.__getAllFoldersFromCache()
        if not folders:
            return []
        return self.__getFoldersForRule(folders, rule)

    def addDatasetEntries(self, datasetId: str, entries):
        entriesList: List[DatasetEntry] = []
        for entry in entries:
            try:
                entriesList.append(DatasetEntry(**entry))
            except Exception as e:
                raise ValueError(
                    f'Couldn\'t cast {entry} to DatasetEntry object. Please adhere to the format of DatasetEntry.')
        MaximAPI.addDatasetEntries(
            self.baseUrl, self.apiKey, datasetId, entriesList)

    def logger(self, config: LoggerConfig):
        # Checking if this log repository exist on server        
        exists = MaximAPI.doesLogRepositoryExist(
            self.baseUrl, self.apiKey, config.id)
        if not exists:
            if config.id:
                raise Exception("Log repository not found")
        if config.id in self.__loggers:
            return self.__loggers[config.id]
        logger = Logger(config=config, api_key=self.apiKey,
                        base_url=self.baseUrl, is_debug=self.__isDebug)
        self.__loggers[config.id] = logger
        return logger

    def cleanup(self):
        self.is_running = False
        self.__syncThread.cancel()
        # Cleaning up all loggers
        for logger in self.__loggers.values():
            logger.cleanup()



import logging
import os
import tempfile
import threading
import time
import uuid
from datetime import datetime, timezone
from queue import Queue
from typing import Optional, Union

from ..apis.maxim import MaximAPI

maximLogger = logging.getLogger("MaximSDK")


class LogWriterConfig:
    def __init__(self, base_url, api_key, repository_id, auto_flush=True, flush_interval: Optional[int] = 10, is_debug=False):
        self.base_url = base_url
        self.api_key = api_key
        self.repository_id = repository_id
        self.auto_flush = auto_flush
        self.flush_interval = flush_interval
        self.is_debug = is_debug
        maximLogger.setLevel(logging.DEBUG if is_debug else logging.INFO)


class LogWriter:
    def __init__(self, config: LogWriterConfig):
        self.is_running = True
        self.id = str(uuid.uuid4())
        self.config = config
        self.queue = Queue()
        self.mutex = threading.Lock()
        self.is_debug = config.is_debug
        self.logs_dir = os.path.join(
            tempfile.gettempdir(), f"maxim-sdk/{self.id}/maxim-logs")
        self.__flush_thread = None
        os.makedirs(self.logs_dir, exist_ok=True)
        if self.config.auto_flush:
            if self.config.flush_interval:
                maximLogger.debug(
                    f"Starting flush thread with interval {self.config.flush_interval} seconds")
                self.__flush_thread = threading.Timer(
                    int(self.config.flush_interval), self.flush)
                self.__flush_thread.start()
            else:
                raise ValueError(
                    "flush_interval is set to None.flush_interval has to be a number")

    def write_to_file(self, logs):
        filename = f"logs-{time.strftime('%Y-%m-%dT%H:%M:%SZ')}.log"
        filepath = os.path.join(self.logs_dir, filename)
        maximLogger.debug(f"Writing logs to file: {filename}")
        with open(filepath, 'w') as file:
            for log in logs:
                file.write(log.serialize() + "\n")
        return filepath

    def flush_log_files(self):
        if os.path.exists(self.logs_dir) == False:
            return
        files = os.listdir(self.logs_dir)
        for file in files:
            with open(os.path.join(self.logs_dir, file), 'r') as f:
                logs = f.read()
            try:
                MaximAPI.pushLogs(
                    self.config.base_url, self.config.api_key, self.config.repository_id, logs)
                os.remove(os.path.join(self.logs_dir, file))
            except Exception as e:
                if self.is_debug:
                    raise Exception(e)

    def flush_logs(self, logs):
        try:
            # Pushing old logs first
            self.flush_log_files()
            # Pushing new logs

            logs_to_push = "\n".join(
                [log.serialize() for log in logs])
            MaximAPI.pushLogs(self.config.base_url, self.config.api_key,
                              self.config.repository_id, logs_to_push)
        except Exception as e:
            self.write_to_file(logs)
            maximLogger.error(
                f"Failed to push logs to server. Writing logs to file. Error: {e}")

    def commit(self, log):
        self.queue.put(log)

    def flush(self):
        if self.is_running == False:
            return
        with self.mutex:
            # Scheduling next call
            if self.config.flush_interval:
                self.__flush_thread = threading.Timer(
                    int(self.config.flush_interval), self.flush)
                self.__flush_thread.start()
            items = []
            while not self.queue.empty():
                items.append(self.queue.get())
            if len(items) == 0:
                maximLogger.debug("No logs to flush")
                return
            maximLogger.debug(
                f"Flushing logs to server {time.strftime('%Y-%m-%dT%H:%M:%S')}")
            if self.is_debug:
                for item in items:
                    maximLogger.debug(f"{item.serialize()}")
        self.flush_logs(items)
        maximLogger.debug("Flush complete")

    def cleanup(self):
        self.flush()
        self.is_running = False
        if self.__flush_thread:
            self.__flush_thread.cancel()
            self.__flush_thread = None



import http.client
import json
import logging
from typing import List, Optional

from ..models.dataset import DatasetEntry
from ..models.folder import Folder
from ..models.prompt import (Prompt, VersionAndRulesWithPromptId,
                             VersionsAndRules)


class MaximAPI:
    @staticmethod
    def __make_network_call(base_url: str, api_key: str, method: str, endpoint: str, body: Optional[str] = None) -> http.client.HTTPResponse:
        if base_url.startswith("http://"):
            conn = http.client.HTTPConnection(base_url.split("//")[1])
        else:
            conn = http.client.HTTPSConnection(base_url.split("//")[1])
        headers = {"x-maxim-api-key": api_key}
        conn.request(method, endpoint, body, headers)
        res = conn.getresponse()
        return res

    @staticmethod
    def getPrompt(base_url: str, api_key: str, id: str) -> VersionsAndRules:
        res = MaximAPI.__make_network_call(
            base_url, api_key, "GET", f"/api/sdk/v3/prompts?promptId={id}")
        if res.status != 200:
            raise Exception(
                f"Error while getting prompt: {res.status} - {res.reason}")
        data = res.read()
        logging.debug(data.decode())
        return VersionsAndRules(**json.loads(data.decode()))

    @staticmethod
    def getPrompts(base_url: str, api_key: str) -> List[VersionAndRulesWithPromptId]:
        res = MaximAPI.__make_network_call(
            base_url, api_key, "GET", "/api/sdk/v3/prompts")
        if res.status != 200:
            raise Exception(
                f"Error while getting prompts: {res.status} - {res.reason}")
        data = res.read()
        return [VersionAndRulesWithPromptId.from_dict(data) for data in json.loads(data)['data']]

    @staticmethod
    def getFolder(base_url: str, api_key: str, id: str) -> Folder:
        res = MaximAPI.__make_network_call(
            base_url, api_key, "GET", f"/api/sdk/v3/folders?folderId={id}")
        if res.status != 200:
            raise Exception(f"Error: {res.status} - {res.reason}")
        data = res.read()
        json_response = json.loads(data.decode())
        if 'tags' not in json_response:
            json_response['tags'] = {}
        return Folder(**json_response['data'])

    @staticmethod
    def getFolders(base_url: str, api_key: str) -> List[Folder]:
        res = MaximAPI.__make_network_call(
            base_url, api_key, "GET", "/api/sdk/v3/folders")
        if res.status != 200:
            raise Exception(f"Error: {res.status} - {res.reason}")
        data = res.read()
        json_response = json.loads(data.decode())
        for elem in json_response['data']:
            if 'tags' not in elem:
                elem['tags'] = {}
        return [Folder(**elem) for elem in json_response['data']]

    @staticmethod
    def addDatasetEntries(base_url: str, api_key: str, dataset_id: str, dataset_entries: List[DatasetEntry]) -> dict:
        res = MaximAPI.__make_network_call(base_url, api_key, "POST", "/api/sdk/v3/datasets/entries", json.dumps(
            {"datasetId": dataset_id, "entries": [entry.to_json() for entry in dataset_entries]}))
        if res.status != 200:
            raise Exception(f"Error: {res.status} - {res.reason}")
        response_data = res.read()
        return json.loads(response_data.decode())

    @staticmethod
    def doesLogRepositoryExist(base_url: str, api_key: str, logger_id: str) -> bool:
        try:
            res = MaximAPI.__make_network_call(
                base_url, api_key, "GET", f"/api/sdk/v3/log-repositories?loggerId={logger_id}")
            if res.status != 200:
                raise Exception(f"Error: {res.status} - {res.reason}")
            response_data = res.read()
            json_response = json.loads(response_data.decode())
            if 'error' in json_response:
                return False
            return True
        except Exception as e:
            return False

    @staticmethod
    def pushLogs(base_url: str, api_key: str, repository_id: str, logs: str) -> None:
        try:
            res = MaximAPI.__make_network_call(
                base_url, api_key, "POST", f"/api/sdk/v3/log?id={repository_id}", logs)
            if res.status != 200:
                raise Exception(f"Error: {res.status} - {res.reason}")
            response_data = res.read()
            json_response = json.loads(response_data.decode())
            if 'error' in json_response:
                raise Exception(json_response['error'])
        except Exception as e:
            raise Exception(e)



import logging
import sys
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID, uuid4

# Use LangChain features
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import BaseMessage
from langchain_core.outputs import LLMResult

from ...expiringKeyValueStore import ExpiringKeyValueStore
from ...logger.components.generation import GenerationConfig, GenerationError
from ...logger.components.retrieval import RetrievalConfig
from ...logger.components.trace import TraceConfig
from ...logger.logger import Logger
from .utils import (parse_langchain_llm_error, parse_langchain_llm_result,
                    parse_langchain_messages, parse_langchain_model_parameters,
                    parse_langchain_provider)

logger = logging.getLogger("MaximSDK")

# 20 minutes
DEFAULT_TIMEOUT = 60*20

class MaximLangchainTracer(BaseCallbackHandler):
    """ A callback handler that logs langchain outputs to Maxim logger

    Args:
        logger: Logger: Maxim Logger instance to log outputs
    """

    def __init__(self, logger: Logger) -> None:
        """ Initializes the Langchain Tracer
        Args:
            logger: Logger: Maxim Logger instance to log outputs
        """
        super().__init__()
        self.logger = logger
        self.generations = ExpiringKeyValueStore()
        self.retrievals = ExpiringKeyValueStore()
        self.metadata_store = ExpiringKeyValueStore()

    def _validate_maxim_metadata(self, metadata: Dict[str, Any]):
        if metadata is None:
            return
        id_keys = ['session_id', 'trace_id', 'span_id']
        present_keys = [key for key in id_keys if key in metadata]
        if len(present_keys) > 1:
            raise ValueError(
                f"Multiple keys found in metadata: {present_keys}. You can pass only one of these.")
        valid_keys = ['session_id', 'trace_id', 'span_id', 'generation_tags',
                      'retrieval_tags', 'trace_tags', 'retrieval_name', 'trace_name', 'generation_name']
        invalid_keys = [key for key in metadata if key not in valid_keys]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"Invalid keys found in metadata: {invalid_keys}. Valid keys are {valid_keys}")

    def _get_container(self, run_id: UUID):
        maxim_metadata = self.metadata_store.get(str(run_id))
        if maxim_metadata is not None:
            span_id = maxim_metadata.get("span_id", None)
            if span_id is not None:
                return "span", span_id
            session_id = maxim_metadata.get("session_id", None)
            if session_id is not None:
                return "session", session_id
            trace_id = maxim_metadata.get("trace_id", None)
            if trace_id is not None:
                return "trace", trace_id
        return "local_trace", str(run_id)

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str],     *,
        run_id: UUID,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any
    ) -> Any:
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        messages = parse_langchain_messages(prompts)

        generation_id = str(uuid4())
        generation_name = None
        generation_tags = None

        if metadata is not None:
            self._validate_maxim_metadata(metadata.get("maxim", None))
            self.metadata_store.set(
                str(run_id), metadata.get("maxim", None), DEFAULT_TIMEOUT)
            generation_name = metadata.get(
                "maxim", {}).get("generation_name", None)
            generation_tags = metadata.get(
                "maxim", {}).get("generation_tags", None)

        generation_config = GenerationConfig(
            id=generation_id, name=generation_name, provider=provider, model=model, messages=messages, model_parameters=model_parameters, tags=generation_tags)
        self.generations.set(str(run_id), generation_id, DEFAULT_TIMEOUT)

        container, container_id = self._get_container(run_id)

        if container == "span":
            self.logger.span_generation(container_id, generation_config)
        elif container == "trace":
            self.logger.trace_generation(container_id, generation_config)
        elif container == "session":
            trace_name = None
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.session_trace(
                container_id, TraceConfig(id=str(run_id), name=trace_name))
            trace.generation(generation_config)
        else:
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.trace(TraceConfig(
                id=str(run_id), name=trace_name))
            trace.generation(generation_config)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        run_id = kwargs.get("run_id", None)
        if run_id is not None:
            result = parse_langchain_llm_result(response)
            generation_id = self.generations.get(str(run_id))
            if generation_id is not None:
                container, container_id = self._get_container(run_id)
                self.logger.generation_result(generation_id, result)
                if container == "local_trace":
                    self.logger.trace_end(container_id)

    def on_retriever_start(self, serialized: Dict[str, Any], query: str, *, run_id: UUID, parent_run_id: UUID | None = None, tags: List[str] | None = None, metadata: Dict[str, Any] | None = None, **kwargs: Any) -> Any:
        """Run when Retriever starts running."""
        container, container_id = self._get_container(run_id)
        retrieval_id = str(uuid4())
        retrieval_config = RetrievalConfig(id=retrieval_id)
        if container == "span":
            retrieval = self.logger.span_retrieval(
                container_id, config=retrieval_config)
            retrieval.input(query)
        elif container == "trace":
            retrieval = self.logger.trace_retrieval(
                container_id, config=retrieval_config)
            retrieval.input(query)
        elif container == "session":
            trace = self.logger.session_trace(
                container_id, TraceConfig(id=str(run_id)))
            retrieval = trace.retrieval(retrieval_config)
            retrieval.input(query)
        else:
            retrieval = self.logger.trace_retrieval(
                str(run_id), config=retrieval_config)
            retrieval.input(query)

    def on_retriever_end(self, documents: Sequence[Document], *, run_id: UUID, parent_run_id: UUID | None = None, **kwargs: Any) -> Any:
        trace_id = str(run_id)
        retrieval_id = self.retrievals.get(trace_id)
        if retrieval_id is not None:
            self.logger.retrieval_output(retrieval_id, documents)

    def on_chat_model_start(
        self, serialized: Dict[str, Any], messages: List[List[BaseMessage]], metadata: Optional[Dict[str, Any]] = None, **kwargs: Any
    ) -> Any:
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        maxim_messages = parse_langchain_messages(messages)
        run_id = kwargs.get("run_id", None)
        generation_id = str(uuid4())
        generation_name = None
        generation_tags = None

        if metadata is not None:
            self._validate_maxim_metadata(metadata.get("maxim", None))
            self.metadata_store.set(
                str(run_id), metadata.get("maxim", None), DEFAULT_TIMEOUT)
            generation_name = metadata.get(
                "maxim", {}).get("generation_name", None)
            generation_tags = metadata.get(
                "maxim", {}).get("generation_tags", None)

        generation_config = GenerationConfig(
            id=generation_id, name=generation_name, provider=provider, model=model, messages=maxim_messages, model_parameters=model_parameters, tags=generation_tags)
        self.generations.set(str(run_id), generation_id, DEFAULT_TIMEOUT)

        container, container_id = self._get_container(run_id)

        if container == "span":
            self.logger.span_generation(container_id, generation_config)
        elif container == "trace":
            self.logger.trace_generation(container_id, generation_config)
        elif container == "session":
            trace_name = None
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.session_trace(
                container_id, TraceConfig(id=str(run_id), name=trace_name))
            trace.generation(generation_config)
        else:
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.trace(TraceConfig(
                id=str(run_id), name=trace_name))
            trace.generation(generation_config)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        run_id = kwargs.get("run_id", None)
        if run_id is not None:
            generation_id = self.generations.get(str(run_id))
            if generation_id is not None:
                container, container_id = self._get_container(run_id)
                generation_error = parse_langchain_llm_error(error)
                self.logger.generation_error(generation_id, generation_error)
                if container == "local_trace":
                    self.logger.trace_end(container_id)

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        logger.error("We don't support chain flows yet")

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        logger.error("Chains are not supported")

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        logger.error("Chains are not supported")

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        logger.error("Tools are not yet supported")

    def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
        logger.error("Tools are not yet supported")

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        logger.error("Tools are not yet supported")

    def on_text(self, text: str, **kwargs: Any) -> Any:
        logger.error("Text are not supported")



import logging
import re
import time
from typing import Any, Dict, List, Union
from uuid import uuid4

from langchain_core.outputs import LLMResult

from ..components.generation import GenerationError

logger = logging.getLogger("MaximSDK")


def parse_langchain_provider(serialized: Dict[str, Any]):
    """ Parses langchain provider from serialized data
    Args:
        serialized: Dict[str, Any]: Serialized data to parse provider from
    Returns:
        str: Parsed provider
    """
    provider = serialized.get("name", "").lower()
    if provider.startswith("chat"):
        return provider.replace("chat", "")
    return provider


def parse_langchain_llm_error(error: Union[Exception, KeyboardInterrupt]) -> GenerationError:
    """ Parses langchain LLM error into a format that is accepted by Maxim logger
    Args:
        error: Union[Exception, KeyboardInterrupt]: Error to be parsed
    Returns:
        GenerationError: Parsed LLM error
    """
    if isinstance(error, KeyboardInterrupt):
        return GenerationError(message="Generation was interrupted by the user")
    else:
        message = error.__dict__.get("message", "")
        type = error.__dict__.get("type", None)
        code = error.__dict__.get("code", None)
        return GenerationError(message=message, type=type, code=code)


def parse_langchain_model_parameters(**kwargs: Any):
    """ Parses langchain kwargs into model and model parameters. You can use this function with any langchain _start callback function
    Args:
        kwargs: Dict[str, Any]: Kwargs to be parsed
    Returns:
        Tuple[str, Dict[str, Any]]: Model and model parameters
    Raises:
        Exception: If model_name is not found in kwargs
    """
    model_parameters = kwargs.get("invocation_params", {})
    model = model_parameters.get("model_name", "")
    del model_parameters["model_name"]
    return model, model_parameters


def parse_langchain_llm_result(result: LLMResult):
    """ Parses langchain LLM result into a format that is accepted by Maxim logger
    Args:
        result: LLMResult: LLM result to be parsed
    Returns:
        Dict[str, Any]: Parsed LLM result
    Raises:
        Exception: If error parsing LLM result
    """
    try:
        generations = result.generations
        choices = []
        if generations is not None:
            for gen_idx, generation in enumerate(generations):
                for idx, gen in enumerate(generation):
                    messages = parse_langchain_messages([gen.text], "system")
                    choices.append({
                        'index': gen_idx+idx,
                        'text': messages[0]['content'],
                        'logprobs': gen.generation_info.get("logprobs") if gen.generation_info else None,
                        'finish_reason': gen.generation_info.get("finish_reason") if gen.generation_info else None
                    })
        usage = result.llm_output.get(
            "token_usage") if result.llm_output else None
        return {
            'id': str(uuid4()),
            'created': int(time.time()),
            'choices': choices,
            'usage': usage
        }
    except Exception as e:
        logger.error(f"Error parsing LLM result: {e}")
        raise Exception(f"Error parsing LLM result: {e}")


def parse_langchain_messages(input: Union[List[str], List[List[Any]]], default_role="user"):
    """ Parses langchain messages into messages that are accepted by Maxim logger
    Args:
        input: List[str] or List[List[Any]]: List of messages to be parsed
        default_role: str: Default role to assign to messages without a role
    Returns:
        List[Dict[str, str]]: List of messages with role and content
    Raises:
        Exception: If input is not List[str] or List[List[Any]]
        Exception: If message type is not str or list
        Exception: If message type is not recognized
    """
    try:
        delimiter_to_role = {
            "System": "system",
            "Human": "user",
            "User": "user",
            "Assistant": "assistant",
            "Model": "model",
        }
        messages = []
        # checking if input is List[str] or List[List]

        if isinstance(input[0], list):
            for message_list in input or []:
                for message in message_list:
                    if isinstance(message, str):
                        continue
                    message_type = type(message).__name__
                    if message_type.endswith("SystemMessage"):
                        messages.append(
                            {"role": "system", "content": message.content or ""})
                    elif message_type.endswith("HumanMessage"):
                        messages.append(
                            {"role": "user", "content": message.content or ""})
                    else:
                        logger.error(
                            f"Invalid message type: {type(message)}")
                        raise Exception(
                            f"Invalid message type: {type(message)}")
        else:
            for message in input or []:
                if not isinstance(message, str):
                    logger.error(f"Invalid message type: {type(message)}")
                    raise Exception(
                        f"Invalid message type: {type(message)}")
                # get type of message
                # Define the delimiter pattern
                pattern = r'(System:|Human:|User:|Assistant:|Model:)'
                # Split the text using the pattern
                splits = re.split(pattern, message)
                # Remove any leading/trailing whitespace and empty strings
                splits = [s.strip() for s in splits if s.strip()]
                # Pair up the delimiters with their content
                for i in range(0, len(splits), 2):
                    if i + 1 < len(splits):
                        # Remove ":" from delimiter and trim both delimiter and content
                        delimiter = splits[i].rstrip(':').strip()
                        content = splits[i+1].strip()
                        messages.append({"role": delimiter_to_role.get(
                            delimiter, "user"), "content": content})
                    else:
                        if splits[i].find(":") == -1:
                            messages.append({"role": delimiter_to_role.get(
                                default_role, "user"), "content": splits[i]})
                        else:
                            # Handle case where there's a delimiter without content
                            delimiter = splits[i].rstrip(':').strip()
                            messages.append({"role": delimiter_to_role.get(
                                delimiter, "user"), "content": ""})
        return messages
    except Exception as e:
        logger.error(f"Error parsing messages: {e}")
        raise Exception(f"Error parsing messages: {e}")



import json
from dataclasses import asdict, dataclass, fields
from datetime import datetime
from typing import Dict, List, Optional, Union


@dataclass
class Message():
    role: str
    content: str

    @staticmethod
    def from_dict(obj: Dict):
        role = obj['role']
        content = obj['content']
        return Message(role=role, content=content)


@dataclass
class Prompt():
    promptId: str
    version: int
    versionId: str
    messages: List[Message]
    modelParameters: Dict[str, Union[str, int, bool, Dict, None]]


@dataclass
class RuleType():
    field: str
    value: Union[str, int, List[str], bool, None]  # adding None here
    operator: str
    valueSource: Optional[str] = None
    exactMatch: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Dict):
        return RuleType(field=obj['field'], value=obj['value'], operator=obj['operator'], valueSource=obj.get('valueSource', None), exactMatch=obj.get('exactMatch', None))


@dataclass
class RuleGroupType():
    rules: List[Union['RuleType', 'RuleGroupType']]
    combinator: str

    @staticmethod
    def from_dict(obj: Dict):
        rules = []
        for rule in obj['rules']:
            if 'rules' in rule:
                rules.append(RuleGroupType.from_dict(rule))
            else:
                rules.append(RuleType(**rule))
        return RuleGroupType(rules=rules, combinator=obj['combinator'])


@dataclass
class PromptDeploymentRules():
    version: int
    query: Optional[RuleGroupType] = None

    @staticmethod
    def from_dict(obj: Dict):
        query = obj.get('query', None)
        if query:
            query = RuleGroupType.from_dict(query)
        return PromptDeploymentRules(version=obj['version'], query=query)


@dataclass
class VersionSpecificDeploymentConfig():
    id: str
    timestamp: datetime
    rules: PromptDeploymentRules
    isFallback: bool = False

    @staticmethod
    def from_dict(obj: Dict):
        rules = PromptDeploymentRules.from_dict(obj['rules'])
        return VersionSpecificDeploymentConfig(id=obj['id'], timestamp=obj['timestamp'], rules=rules, isFallback=obj.get('isFallback', False))


@dataclass
class PromptVersionConfig():
    messages: List[Message]
    modelParameters: Dict[str, Union[str, int, bool, Dict, None]]
    model: str
    tags: Optional[Dict[str, Union[str, int, bool, None]]] = None

    @staticmethod
    def from_dict(obj: Dict):
        messages = [Message.from_dict(message) for message in obj['messages']]
        return PromptVersionConfig(messages=messages, modelParameters=obj['modelParameters'], model=obj['model'], tags=obj.get('tags', None))


@dataclass
class PromptVersion():
    id: str
    version: int
    promptId: str
    createdAt: str
    updatedAt: str
    deletedAt: Optional[str] = None
    description: Optional[str] = None
    config: Optional[PromptVersionConfig] = None

    @staticmethod
    def from_dict(obj: Dict):
        config = obj.get('config', None)
        if config:
            config = PromptVersionConfig.from_dict(config)
        return PromptVersion(id=obj['id'], version=obj['version'], promptId=obj['promptId'], createdAt=obj['createdAt'], updatedAt=obj['updatedAt'], deletedAt=obj.get('deletedAt', None), description=obj.get('description', None), config=config)


@dataclass
class VersionsAndRules():
    rules: Dict[str, List[VersionSpecificDeploymentConfig]]
    versions: List[PromptVersion]
    folderId: Optional[str] = None
    fallbackVersion: Optional[PromptVersion] = None

    @staticmethod
    def from_dict(obj: Dict):
        rules = obj['rules']
        # Decoding each rule
        for key in rules:
            rules[key] = [VersionSpecificDeploymentConfig.from_dict(
                rule) for rule in rules[key]]
        versions = [PromptVersion.from_dict(version)
                    for version in obj['versions']]
        fallbackVersion = obj.get('fallbackVersion', None)
        if fallbackVersion:
            fallbackVersion = PromptVersion.from_dict(fallbackVersion)
        return VersionsAndRules(rules=rules, versions=versions,  folderId=obj.get('folderId', None), fallbackVersion=fallbackVersion)


@ dataclass
class VersionAndRulesWithPromptId(VersionsAndRules):
    promptId: str = ""

    @ staticmethod
    def from_dict(obj: Dict):
        rules = obj['rules']
        # Decoding each rule
        for key in rules:
            rules[key] = [VersionSpecificDeploymentConfig(
                **rule) for rule in rules[key]]
        return VersionAndRulesWithPromptId(rules=rules, versions=obj['versions'], promptId=obj['promptId'], folderId=obj.get('folderId', None), fallbackVersion=obj.get('fallbackVersion', None))


class VersionAndRulesWithPromptIdEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, VersionAndRulesWithPromptId):
            return asdict(obj)
        return super().default(obj)


@ dataclass
class Error():
    message: str


@ dataclass
class PromptData():
    promptId: str
    rules: Dict[str, List[VersionSpecificDeploymentConfig]]
    versions: List[PromptVersion]
    folderId: Optional[str] = None
    fallbackVersion: Optional[PromptVersion] = None


@ dataclass
class MaximApiPromptResponse():
    data: VersionsAndRules
    error: Optional[Error] = None


@ dataclass
class MaximApiPromptsResponse():
    data: List[PromptData]
    error: Optional[Error] = None


@ dataclass
class MaximAPIResponse():
    error: Optional[Error] = None
